#!/usr/bin/env python3

import os
import random
import unittest
from math import pi
from test._utils import least_used_cuda_device

import gpytorch
import torch
from gpytorch.distributions import MultitaskMultivariateNormal
from gpytorch.kernels import GridInterpolationKernel, MultitaskKernel, RBFKernel
from gpytorch.likelihoods import MultitaskGaussianLikelihood
from gpytorch.means import ConstantMean, MultitaskMean


class MultitaskGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super(MultitaskGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = MultitaskMean(ConstantMean(), num_tasks=2)
        self.data_covar_module = GridInterpolationKernel(RBFKernel(), grid_size=100, num_dims=1)
        self.covar_module = MultitaskKernel(self.data_covar_module, num_tasks=2, rank=1)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultitaskMultivariateNormal(mean_x, covar_x)


class TestKroneckerMultiTaskKISSGPRegression(unittest.TestCase):
    def setUp(self):
        if os.getenv("UNLOCK_SEED") is None or os.getenv("UNLOCK_SEED").lower() == "false":
            self.rng_state = torch.get_rng_state()
            torch.manual_seed(0)
            if torch.cuda.is_available():
                torch.cuda.manual_seed_all(0)
            random.seed(0)

    def tearDown(self):
        if hasattr(self, "rng_state"):
            torch.set_rng_state(self.rng_state)

    def _get_data(self, cuda=False):
        # Simple training data: let's try to learn a sine function
        train_x = torch.linspace(0, 1, 100, device=torch.device("cuda") if cuda else torch.device("cpu"))
        # y1 function is sin(2*pi*x) with noise N(0, 0.04)
        train_y1 = torch.sin(train_x * (2 * pi)) + torch.randn_like(train_x) * 0.1
        # y2 function is cos(2*pi*x) with noise N(0, 0.04)
        train_y2 = torch.cos(train_x * (2 * pi)) + torch.randn_like(train_x) * 0.1
        # Create a train_y which interleaves the two
        train_y = torch.stack([train_y1, train_y2], -1)
        return train_x, train_y

    def test_multitask_gp_mean_abs_error(self, cuda=False):
        train_x, train_y = self._get_data(cuda=cuda)
        likelihood = MultitaskGaussianLikelihood(num_tasks=2)
        model = MultitaskGPModel(train_x, train_y, likelihood)

        if cuda:
            model.cuda()

        # Find optimal model hyperparameters
        model.train()
        likelihood.train()

        # Use the adam optimizer
        optimizer = torch.optim.Adam([{"params": model.parameters()}], lr=0.1)  # Includes GaussianLikelihood parameters

        # "Loss" for GPs - the marginal log likelihood
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, model)

        n_iter = 50
        for _ in range(n_iter):
            # Zero prev backpropped gradients
            optimizer.zero_grad()
            # Make predictions from training data
            # Again, note feeding duplicated x_data and indices indicating which task
            output = model(train_x)
            # TODO: Fix this view call!!
            loss = -mll(output, train_y)
            loss.backward()
            optimizer.step()

        # Test the model
        model.eval()
        likelihood.eval()

        test_x = torch.linspace(0, 1, 51, device=torch.device("cuda") if cuda else torch.device("cpu"))
        test_y1 = torch.sin(test_x * (2 * pi))
        test_y2 = torch.cos(test_x * (2 * pi))
        test_preds = likelihood(model(test_x)).mean
        mean_abs_error_task_1 = torch.mean(torch.abs(test_y1 - test_preds[:, 0]))
        mean_abs_error_task_2 = torch.mean(torch.abs(test_y2 - test_preds[:, 1]))

        self.assertLess(mean_abs_error_task_1.squeeze().item(), 0.05)
        self.assertLess(mean_abs_error_task_2.squeeze().item(), 0.05)

    def test_multitask_gp_mean_abs_error_cuda(self):
        if torch.cuda.is_available():
            with least_used_cuda_device():
                self.test_multitask_gp_mean_abs_error(cuda=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import torch


def inv_softplus(x):
    return torch.log(torch.exp(x) - 1)


def inv_sigmoid(x):
    return torch.log(x / (1 - x))


def _get_inv_param_transform(param_transform, inv_param_transform=None):
    reg_inv_tf = TRANSFORM_REGISTRY.get(param_transform, None)
    if reg_inv_tf is None:
        if inv_param_transform is None:
            raise RuntimeError("Must specify inv_param_transform for custom param_transforms")
        return inv_param_transform
    elif inv_param_transform is not None and reg_inv_tf != inv_param_transform:
        raise RuntimeError("TODO")
    return reg_inv_tf


TRANSFORM_REGISTRY = {
    torch.exp: torch.log,
    torch.nn.functional.softplus: inv_softplus,
    torch.nn.functional.sigmoid: inv_sigmoid,
}



#! /usr/bin/env python3

from abc import ABC, abstractproperty

import torch
from gpytorch.likelihoods import LikelihoodList
from gpytorch.models import GP
from torch.nn import ModuleList


class AbstractModelList(GP, ABC):
    @abstractproperty
    def num_outputs(self):
        """The model's number of outputs"""
        pass

    def forward_i(self, i, *args, **kwargs):
        """Forward restricted to the i-th output only"""
        raise NotImplementedError

    def likelihood_i(self, i, *args):
        """Evaluate likelihood of the i-th output only"""
        raise NotImplementedError


class IndependentModelList(AbstractModelList):
    def __init__(self, *models):
        super().__init__()
        self.models = ModuleList(models)
        for m in models:
            if not hasattr(m, "likelihood"):
                raise ValueError(
                    "IndependentModelList currently only supports models that have a likelihood (e.g. ExactGPs)"
                )
        self.likelihood = LikelihoodList(*[m.likelihood for m in models])

    @property
    def num_outputs(self):
        return len(self.models)

    def forward_i(self, i, *args, **kwargs):
        return self.models[i].forward(*args, **kwargs)

    def likelihood_i(self, i, *args):
        return self.likelihood.likelihoods[i](*args)

    def forward(self, *args, **kwargs):
        [model.forward(*args_, **kwargs) for model, args_ in zip(self.models, _get_tensor_args(*args))]

    def __call__(self, *args, **kwargs):
        return [model.__call__(*args_, **kwargs) for model, args_ in zip(self.models, _get_tensor_args(*args))]

    @property
    def train_inputs(self):
        return [model.train_inputs for model in self.models]

    @property
    def train_targets(self):
        return [model.train_targets for model in self.models]


def _get_tensor_args(*args):
    for arg in args:
        if torch.is_tensor(arg):
            yield (arg,)
        else:
            yield arg



#!/usr/bin/env python3

import unittest
from test.means._base_mean_test_case import BaseMeanTestCase
from gpytorch.means import ZeroMean


class TestZeroMean(BaseMeanTestCase, unittest.TestCase):
    def create_mean(self):
        return ZeroMean()



#!/usr/bin/env python3

import math
import torch
from .kernel import Kernel
from ..lazy import delazify, DiagLazyTensor, MatmulLazyTensor, RootLazyTensor, PsdSumLazyTensor
from ..distributions import MultivariateNormal
from ..mlls import InducingPointKernelAddedLossTerm
from ..utils.cholesky import psd_safe_cholesky


class InducingPointKernel(Kernel):
    def __init__(self, base_kernel, inducing_points, likelihood, active_dims=None):
        super(InducingPointKernel, self).__init__(active_dims=active_dims)
        self.base_kernel = base_kernel
        self.likelihood = likelihood

        if inducing_points.ndimension() == 1:
            inducing_points = inducing_points.unsqueeze(-1)
        if inducing_points.ndimension() != 2:
            raise RuntimeError("Inducing points should be 2 dimensional")
        self.register_parameter(name="inducing_points", parameter=torch.nn.Parameter(inducing_points))
        self.register_added_loss_term("inducing_point_loss_term")

    def train(self, mode=True):
        if hasattr(self, "_cached_kernel_mat"):
            del self._cached_kernel_mat
        return super(InducingPointKernel, self).train(mode)

    @property
    def _inducing_mat(self):
        if not self.training and hasattr(self, "_cached_kernel_mat"):
            return self._cached_kernel_mat
        else:
            res = delazify(self.base_kernel(self.inducing_points, self.inducing_points))
            if not self.training:
                self._cached_kernel_mat = res
            return res

    @property
    def _inducing_inv_root(self):
        if not self.training and hasattr(self, "_cached_kernel_inv_root"):
            return self._cached_kernel_inv_root
        else:
            chol = psd_safe_cholesky(self._inducing_mat, upper=True)
            eye = torch.eye(chol.size(-1), device=chol.device)
            inv_root = torch.trtrs(eye, chol)[0]

            res = inv_root
            if not self.training:
                self._cached_kernel_inv_root = res
            return res

    def _get_covariance(self, x1, x2):
        k_ux1 = delazify(self.base_kernel(x1, self.inducing_points))
        if torch.equal(x1, x2):
            covar = RootLazyTensor(k_ux1.matmul(self._inducing_inv_root))

            # Diagonal correction for predictive posterior
            correction = (self.base_kernel(x1, x2, diag=True) - covar.diag()).clamp(0, math.inf)
            covar = PsdSumLazyTensor(covar, DiagLazyTensor(correction))
        else:
            k_ux2 = delazify(self.base_kernel(x2, self.inducing_points))
            covar = MatmulLazyTensor(
                k_ux1.matmul(self._inducing_inv_root), k_ux2.matmul(self._inducing_inv_root).transpose(-1, -2)
            )

        return covar

    def _covar_diag(self, inputs):
        if inputs.ndimension() == 1:
            inputs = inputs.unsqueeze(1)

        # Get diagonal of covar
        covar_diag = delazify(self.base_kernel(inputs, diag=True))
        return DiagLazyTensor(covar_diag)

    def forward(self, x1, x2, diag=False, **kwargs):
        covar = self._get_covariance(x1, x2)

        if self.training:
            if not torch.equal(x1, x2):
                raise RuntimeError("x1 should equal x2 in training mode")
            zero_mean = torch.zeros_like(x1.select(-1, 0))
            new_added_loss_term = InducingPointKernelAddedLossTerm(
                MultivariateNormal(zero_mean, self._covar_diag(x1)),
                MultivariateNormal(zero_mean, covar),
                self.likelihood,
            )
            self.update_added_loss_term("inducing_point_loss_term", new_added_loss_term)

        if diag:
            return covar.diag()
        else:
            return covar

    def num_outputs_per_input(self, x1, x2):
        return self.base_kernel.num_outputs_per_input(x1, x2)



#!/usr/bin/env python3

from torch.autograd import Function
from ..utils.sparse import bdsmm


class DSMM(Function):
    def __init__(self, sparse):
        self.sparse = sparse

    def forward(self, dense):
        return bdsmm(self.sparse, dense)

    def backward(self, grad_output):
        return bdsmm(self.sparse.transpose(-1, -2), grad_output)



#!/usr/bin/env python3

import torch
from torch.autograd import Function
import math


class NormalCDF(Function):
    def __init__(self):
        self.a_for_erf = 8.0 / (3.0 * math.pi) * (math.pi - 3.0) / (4.0 - math.pi)

    def erf_approx(self, x):
        exp = -x * x * (4 / math.pi + self.a_for_erf * x * x) / (1 + self.a_for_erf * x * x)
        return torch.sign(x) * torch.sqrt(1 - torch.exp(exp))

    def erfinv_approx(self, x):
        b = -2 / (math.pi * self.a_for_erf) - torch.log(1 - x * x) / 2
        return torch.sign(x) * torch.sqrt(b + torch.sqrt(b * b - torch.log(1 - x * x) / self.a_for_erf))

    def forward(self, x):
        return (1 + self.erf_approx(x.mul(math.sqrt(0.5)))) / 2

    def backward(self, x):
        """
        The derivative of the standard Normal CDF is the standard normal PDF
        """
        normalizing_constant = 1.0 / math.sqrt(2 * math.pi)
        return x.pow(2).div_(-2).exp_().mul_(normalizing_constant)
